# -*- coding: utf-8 -*-

# Copyright © Cloud Linux GmbH & Cloud Linux Software, Inc 2010-2021 All Rights Reserved
#
# Licensed under CLOUD LINUX LICENSE AGREEMENT
# http://cloudlinux.com/docs/LICENSE.TXT

"""
This module contains X Ray Smart Advice local utility main class
"""
import json
import logging
import os
from contextlib import contextmanager
from dataclasses import asdict
from typing import Any, Tuple, Optional

from schema import Schema, SchemaError

from .advice_types import advice_mapping
from .progress import SmartAdviceProgress
from .schemas import (
    advice_list_schema, detailed_advice_schema, user_sites_info_schema
)
from ..apiclient import get_client
from ..internal.constants import advice_pending_storage, advice_applied_storage
from ..internal.exceptions import XRayAPIError, XRayError
from ..internal.user_plugin_utils import user_mode_advice_verification, username_verification
from ..internal.utils import timestamp, is_wpos_available, safe_move


class SmartAdviceUtil:
    """Main Smart Advice local utility class"""

    def __init__(self):
        if not is_wpos_available():
            raise XRayError('Feature is currently disabled. '
                            'Please, refer to https://docs.cloudlinux.com/cloudlinux-os-plus/#x-ray-smart-advice')
        self.logger = logging.getLogger('smart_advice')
        # check existence of pending and apply storage and create it if missing
        for stor in (advice_pending_storage, advice_applied_storage):
            self.create_dir(stor)
        # initialize Adviser API client
        adviser_client_object = get_client('adviser')
        self.adviser_client = adviser_client_object()

    @staticmethod
    def create_dir(dpath: str) -> None:
        """Create dir if missing"""
        if not os.path.isdir(dpath):
            os.mkdir(dpath)

    @staticmethod
    def response(**kwargs) -> str:
        """
        Create JSON response message with result field == success
        and given keyword arguments in other fields
        :return: json packed string
        """
        initial = {'result': 'success', 'timestamp': timestamp()}
        if kwargs:
            initial.update(kwargs)
        return json.dumps(dict(sorted(initial.items())))

    @staticmethod
    def _apply_datafile(a_id: int) -> str:
        """Per-advice file with results of apply"""
        return f'{advice_applied_storage}/{a_id}'

    def _apply_results(self, a_id: int) -> Optional[str]:
        """Retrieve data stored in per-advice file with results of apply"""
        datafile = self._apply_datafile(a_id)
        if os.path.isfile(datafile):
            try:
                with open(datafile) as _f:
                    data = json.load(_f)
            except OSError:
                data = None
            except json.JSONDecodeError:
                data = _f.read()
            finally:
                os.unlink(datafile)
            return data

    @staticmethod
    def _pending_flag(a_id: int) -> str:
        """Per-advice pending flag"""
        return f'{advice_pending_storage}/{a_id}'

    def is_advice_pending(self, advice_id: int) -> bool:
        """Is advice in pending state"""
        return os.path.exists(self._pending_flag(advice_id))

    def _progress_file(self, a_id: int) -> str:
        """Per-advice progress storage"""
        return self._pending_flag(a_id) + '.progress'

    @staticmethod
    def get_advice_instance(advice_type: str) -> object:
        """Get appropriate instance of Apply actions class by advice type"""
        return advice_mapping[advice_type]

    def _validate(self, data: Any, schema: Schema) -> Any:
        """Validate given data using given schema"""
        try:
            return schema.validate(data)
        except SchemaError as e:
            self.logger.error('Failed to validate API response: %s', data)
            msg = e.errors[-1] or e.autos[-1]
            raise XRayAPIError(f'Malformed API response: {msg}')

    @user_mode_advice_verification
    def get_detailed_advice(self, advice_id: int) -> Tuple[dict, object]:
        """
        Get advice details from API along with an
        appropriate Advice object by obtained advice type
        """
        _response = self._validate(
            data=self.adviser_client.advice_details(advice_id),
            schema=detailed_advice_schema)
        return _response, self.get_advice_instance(_response['advice']['type'])

    @staticmethod
    def dump_to_file(dst: str, data: Any, as_json: bool = False) -> None:
        """Dump data inside given dst using .bkp file and then move"""
        _tmp_dst = dst + '.bkp'
        with open(_tmp_dst, 'w') as _f:
            if as_json:
                try:
                    json.dump(data, _f)
                except TypeError:
                    _f.write(data)
            else:
                _f.write(data)
        safe_move(_tmp_dst, dst)

    def progress(self, advice_id: int, current: SmartAdviceProgress) -> SmartAdviceProgress:
        """
        Smart Advice apply own progress,
        based on current and previous results of 'clwpos get-progress'.
        Returns the maximum progress values among current and previous.
        Previous (latest) is stored inside a pending file-flag.
        """

        @contextmanager
        def resolve_progress() -> SmartAdviceProgress:
            """
            Get progress stored in file and return the maximum one among
            current and stored results
            """
            self.logger.debug('Current progress value: %s', current)
            progress_dst = self._progress_file(advice_id)
            try:
                # read stored result
                with open(progress_dst) as prev:
                    prev_stages = SmartAdviceProgress(**json.load(prev))
            except (OSError, json.JSONDecodeError) as e:
                self.logger.debug('Error during reading stored value: %s',
                                  str(e))
                # or set a dummy one
                prev_stages = SmartAdviceProgress()

            self.logger.debug('Stored progress value: %s', prev_stages)
            yield current if current > prev_stages else prev_stages

            if self.is_advice_pending(advice_id) and current > prev_stages:
                # update stored result only if pending file exists
                self.logger.debug('Updating stored progress: %s', current)
                self.dump_to_file(progress_dst, asdict(current),
                                  as_json=True)

        return resolve_progress()

    def get_current_status(self, _id: int, advice_obj: object,
                           advice_data: dict) -> Tuple[str, dict]:
        """
        Resolve if advice is in pending status.
        Progress is retrieved for pending advice only.
        Dummy result (0, 0) is returned for other advice states.
        """
        if self.is_advice_pending(_id):
            status = 'pending'
            with self.progress(_id, advice_obj.get_progress(
                    advice_data['metadata']['username'])) as p:
                stages = p
        else:
            status = advice_data['advice']['status']
            stages = SmartAdviceProgress()
        return status, asdict(stages)

    def advice_list(self) -> str:
        """Load validated advice list and update it before returning"""
        api_response = self._validate(data=self.adviser_client.advice_list(),
                                      schema=advice_list_schema)
        return self.response(data=self.prepare_advices_response(api_response))

    @username_verification
    def get_site_statuses(self, username) -> str:
        api_response = self._validate(data=self.adviser_client.site_info(username),
                                      schema=user_sites_info_schema)
        site_statuses_data = {}
        for website in api_response:
            domain = website['domain']
            if domain not in site_statuses_data:
                site_statuses_data[domain] = []
            site_statuses_data[domain].append({
                'website': website['website'],
                'urls_scanned': len(website['urls']),
                'advices': self.prepare_advices_response(website['advices'])
            })
        result = [{'domain': domain, 'websites': websites} for domain, websites in site_statuses_data.items()]
        return self.response(data=result)

    def prepare_advices_response(self, from_api):
        advices = from_api.copy()
        for item in advices:
            advice_itself = item['advice']
            advice_type = advice_itself['type']
            advice_instance = self.get_advice_instance(advice_type)
            advice_itself['description'] = advice_instance.short_description

            status, stages = self.get_current_status(advice_itself['id'],
                                                     advice_instance, item)
            advice_itself['status'] = status
            advice_itself.update(stages)
        return advices

    def advice_details(self, advice_id: int) -> str:
        """Load validated advice details and update it before returning"""
        api_response, advice_instance = self.get_detailed_advice(advice_id)

        advice_itself = api_response['advice']

        advice_itself['description'] = advice_instance.short_description
        advice_itself['detailed_description'] = advice_instance.detailed_description
        advice_itself['fix_action'] = advice_instance.apply_button_text

        return self.response(data=api_response)

    def advice_apply(self, advice_id: int, ignore_errors: bool = False,
                     async_mode: bool = False) -> Optional[str]:
        """Apply advice"""
        open(self._pending_flag(advice_id), 'w').close()

        if async_mode:
            # put itself to background in async mode
            # and return current status
            child = os.fork()
            if child:
                return self.advice_apply_status(advice_id)
        return self._exec_advice_apply(advice_id, ignore_errors, async_mode)

    def _exec_advice_apply(self, advice_id: int, ignore_errors: bool = False,
                     async_mode: bool = False) -> Optional[str]:
        """Execute advice apply"""
        try:
            api_response, advice_instance = self.get_detailed_advice(advice_id)

            applied, output = advice_instance.apply(**api_response['metadata'],
                                                    ignore_errors=ignore_errors)
            if applied:
                self.logger.debug('Applied successfully')
                self.adviser_client.update_advice(advice_id=advice_id,
                                                  status='applied')
        finally:
            os.unlink(self._pending_flag(advice_id))

        if async_mode:
            # in async mode write result to a special file
            # and return no output
            self.dump_to_file(self._apply_datafile(advice_id),
                              output)
        else:
            # in sync mode return output
            return output

    def advice_counters(self) -> str:
        """Return advice counters for a server"""
        try:
            api_response = self._validate(data=self.adviser_client.advice_list(),
                                          schema=advice_list_schema)
        except XRayError:
            # bad API responses (500 for example)
            # or JWT token failed check (CL Shared for example)
            result = dict.fromkeys(['total', 'applied'], None)
        else:
            result = dict(
                total=len(api_response),
                applied=len([advice for advice in api_response if
                             advice['advice']['status'] == 'applied'])
            )
        return self.response(data=result)

    def advice_apply_status(self, advice_id: int) -> str:
        """
        Return current status and progress of applying particular advice.
        """
        api_response, advice_instance = self.get_detailed_advice(advice_id)
        status, stages = self.get_current_status(advice_id, advice_instance,
                                                 api_response)
        if status != 'pending':
            # include result of apply for non-pending advice only
            apply_result = self._apply_results(advice_id)
            # drop stored progress
            _progress = self._progress_file(advice_id)
            if os.path.isfile(_progress):
                os.unlink(_progress)

            if apply_result:
                return self.response(status=status, **stages,
                                     data=apply_result)
        return self.response(status=status, **stages)
